import os
from typing import List
import requests
from typing_extensions import Literal, TypedDict, NotRequired
from ai_powered.tool_call import make_tool

class SearchOption(TypedDict):
    siteSearch: NotRequired[str]
    dateRestrict: NotRequired[Literal["d[1]", "w[1]", "m[1]"]]

class SearchResult(TypedDict):
    title: str
    link: str
    snippet: str

GOOGLE_API_KEY = os.getenv("GOOGLE_API_KEY", "mock_google_pse_api_key")
GOOGLE_ENGINE_ID = os.getenv("GOOGLE_ENGINE_ID", "mock_google_pse_engine_id")

@make_tool
def google_search(keywords: str, options: SearchOption) -> List[SearchResult]:
    '''
    Use Google to search the internet.
    For professional inquiries, use English keywords for the search.
    For time-sensitive questions, filter the search results using the dateRestrict option:
        d[1] means within the last day, w[1] means within the last week, m[1] means within the last month.
    '''

    url = "https://www.googleapis.com/customsearch/v1"
    params : dict[str, str] = {
        "key": GOOGLE_API_KEY,
        "cx": GOOGLE_ENGINE_ID,
        "q": keywords,
        **options #type: ignore
    }

    response = requests.get(url, params=params)
    response.raise_for_status()  # if failed, raise exception

    search_results = response.json()
    results : list[SearchResult] = []

    for item in search_results.get("items", []):
        results.append(SearchResult(
            title=item["title"],
            link=item["link"],
            snippet=item["snippet"]
        ))

    return results



import os
from typing import Optional

from ai_powered.llm.definitions import ModelFeature

DEBUG = os.environ.get('DEBUG', 'False').lower() in {'true', '1', 'yes', 'on'}
OPENAI_API_KEY = os.environ.get("OPENAI_API_KEY", "sk-1234567890ab-MOCK-API-KEY")
OPENAI_BASE_URL = os.environ.get("OPENAI_BASE_URL", "https://api.openai.com/v1")
OPENAI_MODEL_NAME = os.environ.get("OPENAI_MODEL_NAME")
_features_str = os.environ.get("OPENAI_MODEL_FEATURES")
OPENAI_MODEL_FEATURES : Optional[set[ModelFeature]] = set(ModelFeature[s] for s in _features_str.split(',')) if _features_str else None

SYSTEM_PROMPT = """
You are a function simulator,
try to understand the intent of the specified function and inference its output for given arguments,
the user message contains function arguments and your response should be the return value.

The function to simulate is:
{signature}
    ''' {docstring} '''

The user message contains arguments conforming following json schema:
{parameters_schema}
"""

SYSTEM_PROMPT_RETURN_SCHEMA = """
Your response should only contains exactly one JSON with valid syntax and conforming to the following JSON Schema, without any other words (ofcourse, you can still think step by step, but just don't say it out):
{return_schema}
"""

SYSTEM_PROMPT_JSON_SYNTAX = """
In case you don't know about the syntax of JSON, here are some example:

- number: 2
- string: "hello"
- list: [1, 2, 3]
- dictionary: {"key": "value"}
- boolean: true or false
- null value: null
"""



from abc import ABC
from dataclasses import dataclass, field
import json
from typing import Any, Iterable, Set
from easy_sync import sync_compatible
import openai
from openai.types.chat.chat_completion_message import ChatCompletionMessage
from openai.types.chat.chat_completion_tool_choice_option_param import ChatCompletionToolChoiceOptionParam
from openai.types.chat.chat_completion import ChatCompletion
from openai.types.chat.completion_create_params import ResponseFormat
from ai_powered.colors import green, red, yellow
from ai_powered.constants import DEBUG, SYSTEM_PROMPT
from ai_powered.llm.connection import LlmConnection
from ai_powered.llm.definitions import FunctionSimulator, ModelFeature
from ai_powered.tool_call import ChatCompletionToolParam

@dataclass
class GenericFunctionSimulator (FunctionSimulator, ABC):
    ''' implementation of FunctionSimulator for OpenAI compatible models '''

    connection: LlmConnection
    model_name: str
    model_features: Set[ModelFeature]
    model_options: dict[str, Any]
    system_prompt : str = field(init=False)
    _param_tools : Iterable[ChatCompletionToolParam] | openai.NotGiven = field(init=False)
    _param_tool_choice : ChatCompletionToolChoiceOptionParam | openai.NotGiven = field(init=False)
    _param_response_format : ResponseFormat | openai.NotGiven = field(init=False)

    def __post_init__(self):
        self.system_prompt = self._system_prompt_maker()
        self._param_tools = self._param_tools_maker()
        self._param_tool_choice = self._param_tool_choice_maker()
        self._param_response_format = self._param_response_format_maker()

    def _system_prompt_maker(self) -> str:
        ''' default impl is provided '''
        common_system_prompt = SYSTEM_PROMPT.format(
            signature = self.signature,
            docstring = self.docstring or "no doc, guess intention from function name",
            parameters_schema = json.dumps(self.parameters_schema),
        )
        extra_system_prompt = self._extra_system_prompt_maker()
        return common_system_prompt + extra_system_prompt

    def _extra_system_prompt_maker(self) -> str:
        ''' to be overrided '''
        return ""

    def _param_response_format_maker(self) -> ResponseFormat | openai.NotGiven:
        ''' to be overrided '''
        return openai.NOT_GIVEN

    def _param_tools_maker(self) -> Iterable[ChatCompletionToolParam] | openai.NotGiven:
        ''' to be overrided '''
        return openai.NOT_GIVEN

    def _param_tool_choice_maker(self) -> ChatCompletionToolChoiceOptionParam | openai.NotGiven:
        ''' to be overrided '''
        return openai.NOT_GIVEN

    @sync_compatible
    async def _chat_completion_query(self, arguments_json: str) -> ChatCompletion:
        ''' default impl is provided '''
        return await self.connection.chat_completions(
            model = self.model_name,
            messages = [
                {"role": "system", "content": self.system_prompt},
                {"role": "user", "content": arguments_json}
            ],
            tools = self._param_tools,
            tool_choice = self._param_tool_choice,
            response_format=self._param_response_format,
        )

    def _response_message_parser(self, response_message: ChatCompletionMessage) -> str:
        ''' to be overrided '''
        if DEBUG:
            print(red(f"[GenericFunctionSimulator._response_message_parser()] {self.__class__ =}, {self._response_message_parser =}"))
        raise NotImplementedError

    #@override
    @sync_compatible
    async def query_model(self, arguments_json: str) -> str:

        if DEBUG:
            print(yellow(f"{arguments_json =}"))
            print(yellow(f"request.tools = {self._param_tools}"))
            print(green(f"[fn {self.function_name}] request prepared."))

        response = await self._chat_completion_query(arguments_json)

        if DEBUG:
            print(yellow(f"{response =}"))
            print(green(f"[fn {self.function_name}] response received."))

        response_message = response.choices[0].message
        result_str = self._response_message_parser(response_message)
        return result_str



import asyncio
from functools import wraps
from typing import Any, Awaitable, Callable, Generic, overload
from easy_sync import sync_compatible
from typing_extensions import ParamSpec, TypeVar
import json
import msgspec

from ai_powered.llm.connection import LlmConnection
from ai_powered.llm.definitions import ModelFeature
from ai_powered.llm.adapter_selector import FunctionSimulatorSelector
from ai_powered.llm.known_models import complete_model_config
from ai_powered.schema_deref import deref
from ai_powered.constants import DEBUG, OPENAI_API_KEY, OPENAI_BASE_URL, OPENAI_MODEL_FEATURES, OPENAI_MODEL_NAME, SYSTEM_PROMPT, SYSTEM_PROMPT_JSON_SYNTAX, SYSTEM_PROMPT_RETURN_SCHEMA
from ai_powered.colors import gray, green
import inspect

A = TypeVar("A")
class Result (msgspec.Struct, Generic[A]):
    result: A

P = ParamSpec("P")
R = TypeVar("R")

@overload
def ai_powered(fn: Callable[P, Awaitable[R]]) -> Callable[P, Awaitable[R]]:
    ...

@overload
def ai_powered(fn: Callable[P, R]) -> Callable[P, R]:
    ...

def ai_powered(fn : Callable[P, Awaitable[R]] | Callable[P, R]) -> Callable[P, Awaitable[R]] | Callable[P, R]:
    ''' Provide an AI powered implementation of a function '''

    function_name = fn.__name__
    sig = inspect.signature(fn)
    docstring = inspect.getdoc(fn)

    if DEBUG:
        print(f"{sig =}")
        print(f"{docstring =}")

        for param in sig.parameters.values():
            print(f"{param.name}: {param.annotation}")

        print(f"{sig.return_annotation =}")

    parameters_schema = {param.name: msgspec.json.schema(param.annotation) for param in sig.parameters.values()}
    return_type = sig.return_annotation
    raw_return_schema = msgspec.json.schema(return_type)
    return_schema = deref(raw_return_schema)
    result_type = Result[return_type]

    if DEBUG:
        for param_name, schema in parameters_schema.items():
            print(f"{param_name} (json schema): {schema}")
        print(f"return (json schema): {return_schema}")

    connection = LlmConnection(api_key=OPENAI_API_KEY, base_url=OPENAI_BASE_URL)
    model_config = complete_model_config(OPENAI_BASE_URL, OPENAI_MODEL_NAME, OPENAI_MODEL_FEATURES)
    model_name = model_config.model_name
    model_features: set[ModelFeature] = model_config.supported_features
    model_options: dict[str, Any] = model_config.suggested_options

    sys_prompt = SYSTEM_PROMPT.format(
        signature = sig,
        docstring = docstring or "no doc, guess intention from function name",
        parameters_schema = json.dumps(parameters_schema),
    ) + ("" if ModelFeature.tools in model_features else SYSTEM_PROMPT_RETURN_SCHEMA.format(
        return_schema = json.dumps(return_schema),
    ) + SYSTEM_PROMPT_JSON_SYNTAX )

    if DEBUG:
        print(f"{sys_prompt =}")
        print(f"{return_schema =}")

    fn_simulator = FunctionSimulatorSelector(
        function_name, f"{sig}", docstring, parameters_schema, return_schema,
        connection, model_name, model_features, model_options
    )

    if DEBUG:
        print("fn_simulator =", gray(f"{fn_simulator}"))
        print(green(f"[fn {function_name}] AI is powering up."))

    @wraps(fn)
    def wrapper_fn(*args: P.args, **kwargs: P.kwargs) -> R:
        real_arg = sig.bind(*args, **kwargs)
        real_arg_str = msgspec.json.encode(real_arg.arguments).decode('utf-8')

        if DEBUG:
            print(f"{real_arg_str =}")

        resp_str = fn_simulator.query_model(real_arg_str).wait()

        if DEBUG:
            print(f"{resp_str =}")
            print(green(f"[fn {function_name}] response extracted."))

        returned_result = msgspec.json.decode(resp_str, type=result_type)
        if DEBUG:
            print(f"{returned_result =}")
            print(green(f"[fn {function_name}] response validated."))

        return returned_result.result #type: ignore

    @wraps(fn)
    @sync_compatible
    async def wrapper_fn_async(*args: P.args, **kwargs: P.kwargs) -> R:
        real_arg = sig.bind(*args, **kwargs)
        real_arg_str = msgspec.json.encode(real_arg.arguments).decode('utf-8')

        if DEBUG:
            print(f"{real_arg_str =}")

        # NOTE: the main logic
        resp_str = await fn_simulator.query_model(real_arg_str)

        if DEBUG:
            print(f"{resp_str =}")
            print(green(f"[fn {function_name}] response extracted."))

        returned_result = msgspec.json.decode(resp_str, type=result_type)
        if DEBUG:
            print(f"{returned_result =}")
            print(green(f"[fn {function_name}] response validated."))

        return returned_result.result #type: ignore


    if asyncio.iscoroutinefunction(fn):
        return wrapper_fn_async
    else:
        return wrapper_fn



import openai
from openai.types.chat.chat_completion_message import ChatCompletionMessage
from openai.types.chat.completion_create_params import ResponseFormat
from ai_powered.llm.adapters.generic_adapter import GenericFunctionSimulator
#from ai_powered.llm.definitions import ModelFeature

class StructuredOutputFunctionSimulator (GenericFunctionSimulator):
    ''' implementation of FunctionSimulator for OpenAI compatible models '''

    def _param_response_format_maker(self) -> ResponseFormat | openai.NotGiven:
        ''' to be overrided '''
        return {
            "type": "json_schema",
            "json_schema": {
                "name": "math_response",
                #"strict": True if ModelFeature.strict_mode in self.model_features else None, #NOTE: https://platform.openai.com/docs/guides/structured-outputs/supported-schemas
                "schema": {
                    "type": "object",
                    "properties": {
                        "result": self.return_schema,
                    },
                    "required": ["result"],
                }
            }
        }

    def _response_message_parser(self, response_message: ChatCompletionMessage) -> str:
        tool_calls = response_message.tool_calls
        assert tool_calls is None

        raw_resp_str = response_message.content
        assert raw_resp_str is not None

        return raw_resp_str



import re
from typing import Optional

_json_pattern = re.compile(r'({.*})', re.DOTALL)

def extract_json_from_message(message: str) -> Optional[str]:
    reversed_message = message[::-1]

    # 在反转消息中寻找第一个完整的 {} 对
    match = _json_pattern.search(reversed_message)
    if match:
        # 获取反转字符串中匹配的起始和结束索引
        start, end = match.span()

        # 计算原始字符串中的索引
        original_start = len(message) - end
        original_end = len(message) - start

        # 从原始消息中提取 JSON 字符串
        json_str = message[original_start:original_end]

        return json_str
    else:
        return None



from abc import ABC
from dataclasses import dataclass
import enum
from typing import Any, Optional

from easy_sync import sync_compatible

class ModelFeature (enum.Enum):
    '''
    Ollama Doc: https://ollama.fan/reference/openai/#supported-features
    OpenAI Doc:
    - tools/function_calling: https://platform.openai.com/docs/guides/function-calling
    - structured_outputs/strict_mode: https://platform.openai.com/docs/guides/structured-outputs
    - seed/reproducible_outputs: https://platform.openai.com/docs/advanced-usage/reproducible-outputs
    '''
    tools = "function_calling"
    function_calling = "function_calling"
    seed = "reproducible_outputs"
    reproducible_outputs = "reproducible_outputs"
    structured_outputs = "structured_outputs"
    strict_mode = "strict_mode"

ALL_FEATURES : set[ModelFeature] = {x for x in ModelFeature}


@dataclass
class FunctionSimulator (ABC):
    ''' just a wrapper to call model, without checking type correctness '''

    function_name: str
    signature: str
    docstring: Optional[str]
    parameters_schema: dict[str, Any]
    return_schema: dict[str, Any]

    @sync_compatible
    async def query_model(self, arguments_json: str) -> str:
        ...



from dataclasses import dataclass, field
from typing import Any, ClassVar
import openai
from easy_sync import sync_compatible
from openai.types.chat.chat_completion_message_param import ChatCompletionMessageParam

from ai_powered.colors import gray
from ai_powered.constants import DEBUG, OPENAI_API_KEY, OPENAI_BASE_URL, OPENAI_MODEL_NAME, OPENAI_MODEL_FEATURES
from ai_powered.llm.known_models import complete_model_config
from ai_powered.llm.connection import LlmConnection
from ai_powered.tool_call import ChatCompletionToolParam, MakeTool

default_connection = LlmConnection(base_url=OPENAI_BASE_URL, api_key=OPENAI_API_KEY)
model_config = complete_model_config(OPENAI_BASE_URL, OPENAI_MODEL_NAME, OPENAI_MODEL_FEATURES)

@dataclass
class ChatBot:
    ''' A baseclass to define chatbot which can use tools '''

    system_prompts: ClassVar[list[str]] = []
    tools: ClassVar[tuple[MakeTool[..., Any], ...]] = ()
    connection: ClassVar[LlmConnection] = default_connection
    conversation: list[ChatCompletionMessageParam] = field(default_factory = lambda: [])

    def __post_init__(self):
        self._tool_dict = {tool.fn.__name__: tool for tool in self.tools}
        self._tool_schemas: list[ChatCompletionToolParam] | openai.NotGiven = [t.schema() for t in self.tools
                                                                              ] if len(self.tools) > 0 else openai.NOT_GIVEN

    @sync_compatible
    async def chat_continue(self) -> str:
        if DEBUG:
            print(gray(f"{self.conversation =}"))

        # NOTE: system_prompts might be updated dynamically
        self._system_prompts: list[ChatCompletionMessageParam] = [{"role": "system", "content": s} for s in self.system_prompts]

        response = await self.connection.chat_completions(
            model = model_config.model_name,
            messages = [*self._system_prompts, *self.conversation],
            tools = self._tool_schemas,
        )
        assistant_message = response.choices[0].message

        self.conversation.append(assistant_message.to_dict()) #type: ignore

        tool_calls = assistant_message.tool_calls
        if tool_calls is not None:
            if DEBUG:
                print(gray(f"{len(tool_calls) =}"))

            for tool_call in tool_calls:
                using_tool = self._tool_dict[tool_call.function.name]
                function_message = using_tool.call(tool_call) #type: ignore #TODO: async & parrallel
                self.conversation.append(function_message)

            return await self.chat_continue()
        else:
            message_content = assistant_message.content
            assert message_content is not None
            return message_content

    @sync_compatible
    async def chat(self, message: str) -> str:
        self.conversation.append({"role": "user", "content": message})
        return await self.chat_continue()



from typing import Iterable
import openai
from openai.types.chat.chat_completion_message import ChatCompletionMessage
from openai.types.chat.chat_completion_tool_choice_option_param import ChatCompletionToolChoiceOptionParam
from ai_powered.tool_call import ChatCompletionToolParam
from ai_powered.llm.adapters.generic_adapter import GenericFunctionSimulator
from ai_powered.utils.parse_message import extract_json_from_message

class ToolsFunctionSimulator (GenericFunctionSimulator):
    ''' implementation of FunctionSimulator for OpenAI compatible models '''

    def _param_tools_maker(self) -> Iterable[ChatCompletionToolParam] | openai.NotGiven:
        return [{
            "type": "function",
            "function": {
                "name": "return_result",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "result": self.return_schema,
                    },
                    "required": ["result"],
                }
            },
        }]

    def _param_tool_choice_maker(self) -> ChatCompletionToolChoiceOptionParam | openai.NotGiven:
        return {"type": "function", "function": {"name": "return_result"}}

    #@override
    def _response_message_parser(self, response_message: ChatCompletionMessage) -> str:
        tool_calls = response_message.tool_calls

        if tool_calls is not None:
            return tool_calls[0].function.arguments
        else: # 兼容不支持 tool_choice 的情况, 比如 ollama
            message = response_message.content
            assert message is not None

            json_str = extract_json_from_message(message)
            assert json_str is not None

            return json_str



from copy import deepcopy
from dataclasses import dataclass, field
from typing import Any, Callable, Optional, Set
from typing_extensions import TypeAlias

from ai_powered.llm.definitions import ALL_FEATURES, ModelFeature

@dataclass(frozen=True)
class KnownPlatform:
    ''' information about a known platform '''

    platform_name: str
    match_platform_url: Callable[[str], bool]
    known_model_list: list["KnownModel"]

@dataclass(frozen=True)
class KnownModel:
    ''' information about a known model '''

    model_name: str
    supported_features: Set[ModelFeature]
    suggested_options: dict[str, Any] = field(default_factory=dict)

    def copy(self) -> "KnownModel":
        ''' create a copy of the known model information '''
        return KnownModel(
            model_name = self.model_name,
            supported_features = deepcopy(self.supported_features),
            suggested_options = deepcopy(self.suggested_options)
        )

    def override(self, model_name: Optional[str] = None, supported_features: Optional[Set[ModelFeature]] = None, suggested_options: Optional[dict[str, Any]] = None) -> "KnownModel":
        ''' override the known model information '''
        return KnownModel(
            model_name = model_name if model_name is not None else self.model_name,
            supported_features = deepcopy(supported_features or self.supported_features),
            suggested_options = deepcopy(suggested_options or self.suggested_options)
        )

def contains(s: str) -> Callable[[str], bool]:
    ''' create a function to check if a string contains a substring '''
    return lambda text: s in text

def starts_with(s: str) -> Callable[[str], bool]:
    ''' create a function to check if a string starts with a substring '''
    return lambda text: text.startswith(s)

def equals(s: str) -> Callable[[str], bool]:
    ''' create a function to check if a string contains a substring '''
    return lambda text: s == text

KNOWN_PLATFORMS : list[KnownPlatform] = [
    KnownPlatform(
        platform_name = "openai",
        match_platform_url = contains("openai"),
        known_model_list = [
            KnownModel("gpt-4o-mini", ALL_FEATURES),
            KnownModel("gpt-4o", ALL_FEATURES),
        ]
    ),
    KnownPlatform(
        platform_name = "deepseek",
        match_platform_url = contains("deepseek"),
        known_model_list = [
            KnownModel("deepseek-chat", {ModelFeature.tools}),
            KnownModel("deepseek-coder", {ModelFeature.tools}),
        ]
    ),
    KnownPlatform(
        platform_name = "localhost",
        match_platform_url = contains("localhost"),
        known_model_list = [
            KnownModel("gorilla-llm/gorilla-openfunctions-v2-gguf/gorilla-openfunctions-v2-q4_K_M.gguf", ALL_FEATURES),
            KnownModel("lmstudio-community/Meta-Llama-3-8B-Instruct-GGUF/Meta-Llama-3-8B-Instruct-Q4_K_M.gguf", set()),
            KnownModel("lmstudio-community/DeepSeek-Coder-V2-Lite-Instruct-GGUF", set()),
            KnownModel("firefunction-v2:70b-q2_K", {ModelFeature.tools}),
            KnownModel("deepseek-coder-v2", set()),
            KnownModel("llama3.1", set()),
        ]
    ),
]

ModelConfig : TypeAlias = KnownModel

def complete_model_config(platform_url: str, model_name: Optional[str], model_supported_features: Optional[Set[ModelFeature]]) -> ModelConfig:
    ''' select a known model from a known platform '''

    for platform in KNOWN_PLATFORMS:
        if platform.match_platform_url(platform_url):
            #NOTE: known platform

            #NOTE: default for unknown model_name (not specified or not found in known models)
            selected_config = platform.known_model_list[0]

            if model_name is not None:
                for known_model in platform.known_model_list:
                    if model_name.startswith(known_model.model_name):
                        #NOTE: known platform and model_name
                        selected_config = known_model

            return selected_config.override(model_name=model_name, supported_features=model_supported_features)

    #NOTE: unknown platform
    if model_name is not None:
        return ModelConfig(model_name, (model_supported_features or {ModelFeature.function_calling}))
    else:
        raise ValueError(f"Unknown platform: {platform_url}, please set OPENAI_MODEL_NAME manually")



import json
import re
from openai.types.chat.chat_completion_message import ChatCompletionMessage
from ai_powered.constants import DEBUG, SYSTEM_PROMPT_JSON_SYNTAX, SYSTEM_PROMPT_RETURN_SCHEMA
from ai_powered.llm.adapters.generic_adapter import GenericFunctionSimulator
from ai_powered.utils.parse_message import extract_json_from_message


_result_pattern = re.compile(r'^\s*\{\s*"result"\s*:') # "{"result": 2}" | "{ "result" : 2 }"

class ChatFunctionSimulator (GenericFunctionSimulator):
    ''' implementation of FunctionSimulator for OpenAI compatible models '''

    def _extra_system_prompt_maker(self) -> str:
        return SYSTEM_PROMPT_RETURN_SCHEMA.format(
            return_schema = json.dumps(self.return_schema),
        ) + SYSTEM_PROMPT_JSON_SYNTAX

    def _response_message_parser(self, response_message: ChatCompletionMessage) -> str:
        tool_calls = response_message.tool_calls

        if tool_calls is not None:
            return tool_calls[0].function.arguments
        else:
            raw_resp_str = response_message.content
            assert raw_resp_str is not None

            if DEBUG:
                print(f"{raw_resp_str =}")

            json_str = extract_json_from_message(raw_resp_str) or raw_resp_str

            if DEBUG:
                print(f"{json_str =}")

            if re.match(_result_pattern, json_str): # wrapped by `{"result": ...}`
                result_str = json_str
            else: # not wrapped by `{"result": ...}`
                result_str = f'{{"result": {json_str}}}'

            if DEBUG:
                print(f"{result_str =}")

            return result_str



from typing import Any, Callable, TypeVar
from typing_extensions import ParamSpec


P = ParamSpec('P')
R = TypeVar('R')


def wraps(origin_fn: Callable[P, R]) -> Callable[ [Callable[..., Any]], Callable[P, R]]:
    def wrapper(fn: Callable[..., Any]) -> Callable[P, R]:
        return fn #type: ignore
    return wrapper


def wraps_arguments_type(origin_fn: Callable[P, Any]) -> Callable[ [Callable[..., R]], Callable[P, R]]:
    def wrapper(fn: Callable[..., R]) -> Callable[P, R]:
        return fn #type: ignore
    return wrapper




from typing import Any

def deref(schema: dict[str, Any]) -> dict[str, Any]:
    '''
    msgspec 给出的 json schema 会包含 $ref 字段，这个函数用于将 $ref 字段替换为具体的 schema

    这一行为是符合标准的，见:  https://json-schema.org/understanding-json-schema/structuring#dollarref

    但 OpenAI 及大部分 LLM 服务商尚不支持 $ref 字段，所以需要将其展开以获得更好的效果.

    当然这一做法也有其局限性，比如无法处理环状引用，但这只能等到服务商支持 $ref 字段后再解决.

    {'$ref': '#/$defs/ArgObj', '$defs': {'ArgObj': {'title': 'ArgObj', 'type': 'object', 'properties': {'python_expression': {'type': 'string'}}, 'required': ['python_expression']}}}

    最终返回结果的最外层形如:

    return {
        "type": "object",
        "properties": properties,
        "required": [param.name for param in sig.parameters.values() if param.default == inspect.Parameter.empty],
    }
    '''

    #TODO: 环状引用检测（递归类型，如 Nat，List）

    defs = schema.get("$defs", {})

    def _deref(schema: dict[str, Any]) -> dict[str, Any]:
        if '$ref' in schema:
            ref_name = schema["$ref"].split('/')[-1]
            return _deref(defs[ref_name])
        elif 'properties' in schema:
            properties = {key: _deref(value) for key, value in schema['properties'].items()}
            return schema | {"properties": properties}
        else:
            return schema

    return _deref(schema)



from easy_sync import sync_compatible
from ai_powered.llm.adapters.generic_adapter import GenericFunctionSimulator
from ai_powered.llm.adapters.tools_adapter import ToolsFunctionSimulator
from ai_powered.llm.adapters.chat_adapter import ChatFunctionSimulator
from ai_powered.llm.adapters.structured_output_adapter import StructuredOutputFunctionSimulator
from ai_powered.llm.definitions import ModelFeature


class FunctionSimulatorSelector (GenericFunctionSimulator):
    ''' implementation of FunctionSimulator for OpenAI compatible models '''

    _selected_impl: GenericFunctionSimulator

    def _select_impl(self) -> GenericFunctionSimulator:
        if ModelFeature.structured_outputs in self.model_features:
            return StructuredOutputFunctionSimulator(
                self.function_name, self.signature, self.docstring, self.parameters_schema, self.return_schema,
                self.connection, self.model_name, self.model_features, self.model_options
            )
        elif ModelFeature.tools in self.model_features:
            return ToolsFunctionSimulator(
                self.function_name, self.signature, self.docstring, self.parameters_schema, self.return_schema,
                self.connection, self.model_name, self.model_features, self.model_options
            )
        else:
            return ChatFunctionSimulator(
                self.function_name, self.signature, self.docstring, self.parameters_schema, self.return_schema,
                self.connection, self.model_name, self.model_features, self.model_options
            )

    def __post_init__(self):
        super().__post_init__()
        self._selected_impl = self._select_impl()

    @sync_compatible
    async def query_model(self, arguments_json: str) -> str:
        return await self._selected_impl.query_model(arguments_json)



from functools import partial
from typing import Any, Mapping, Union
from easy_sync import Waitable, sync_compatible
import httpx
import openai
from openai.types.chat.chat_completion import ChatCompletion
from ai_powered.utils.function_wraps import wraps_arguments_type


class LlmConnection:
    sync_client: openai.OpenAI
    async_client: openai.AsyncOpenAI
    base_url: str | httpx.URL | None

    def __init__(self,
        api_key: str | None = None,
        organization: str | None = None,
        project: str | None = None,
        base_url: str | httpx.URL | None = None,
        timeout: Union[float, httpx.Timeout, None, openai.NotGiven] = openai.NOT_GIVEN,
        max_retries: int = openai.DEFAULT_MAX_RETRIES,
        default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        sync_http_client: httpx.Client | None = None,
        async_http_client: httpx.AsyncClient | None = None,
    ):
        self.base_url = base_url

        self.sync_client = openai.OpenAI(
            api_key=api_key,
            organization=organization,
            project=project,
            base_url=base_url,
            timeout=timeout,
            max_retries=max_retries,
            default_headers=default_headers,
            default_query=default_query,
            http_client=sync_http_client,
        )

        self.async_client = openai.AsyncOpenAI(
            api_key=api_key,
            organization=organization,
            project=project,
            base_url=base_url,
            timeout=timeout,
            max_retries=max_retries,
            default_headers=default_headers,
            default_query=default_query,
            http_client=async_http_client,
        )

        async_fn = partial(self.async_client.chat.completions.create, stream=False)
        sync_fn = partial(self.sync_client.chat.completions.create, stream=False)

        f = sync_compatible(sync_fn=sync_fn)(async_fn) #type: ignore
        self._chat_completions = f

    @wraps_arguments_type(openai.AsyncOpenAI().chat.completions.create)
    def chat_completions(self, *args: list[Any], **kwargs: dict[str, Any]) -> Waitable[ChatCompletion]:
        return self._chat_completions(*args, **kwargs)



import inspect
from typing import Callable, Generic
from typing_extensions import ParamSpec, TypeVar
import msgspec

from openai.types.chat.chat_completion_tool_param import ChatCompletionToolParam
from openai.types.chat.chat_completion_named_tool_choice_param import ChatCompletionNamedToolChoiceParam
from openai.types.chat.chat_completion_tool_message_param import ChatCompletionToolMessageParam
from openai.types.chat.chat_completion_message_tool_call import ChatCompletionMessageToolCall
from ai_powered.colors import gray, green
from ai_powered.constants import DEBUG
from ai_powered.schema_deref import deref

P = ParamSpec("P")
R = TypeVar("R")

class MakeTool(msgspec.Struct, Generic[P, R]):
    fn : Callable[P, R]

    def __call__(self, *args: P.args, **kwds: P.kwargs) -> R:
        return self.fn(*args, **kwds)

    def struct_of_parameters(self) -> type[msgspec.Struct]:
        ''' Transmission object corresponding to function parameters '''
        sig = inspect.signature(self.fn)
        properties = [(param.name, param.annotation) for param in sig.parameters.values()]
        ArgObj = msgspec.defstruct("ArgObj", properties)
        return ArgObj

    def schema(self) -> ChatCompletionToolParam:
        function_name = self.fn.__name__
        sig = inspect.signature(self.fn)
        docstring = inspect.getdoc(self.fn)

        if DEBUG:
            print(gray(f"[MakeTool.schema()] {sig =}"))

        raw_schema = msgspec.json.schema(self.struct_of_parameters())
        parameters_schema = deref(raw_schema)

        if DEBUG:
            print(gray(f"[MakeTool.schema()] {parameters_schema =}"))

        return {
            "type": "function",
            "function": {
                "name": function_name,
                "description": docstring or "",
                "parameters": parameters_schema,
            }
        }

    def call(self, tool_call: ChatCompletionMessageToolCall) -> ChatCompletionToolMessageParam:
        if DEBUG:
            print(green(f"[MakeTool.call()] {tool_call =}"))

        assert self.fn.__name__ == tool_call.function.name

        sig = inspect.signature(self.fn)
        args_json : str = tool_call.function.arguments
        args_obj = msgspec.json.decode(args_json, type=self.struct_of_parameters())
        args_dict = msgspec.structs.asdict(args_obj)

        if DEBUG:
            print(gray(f"[MakeTool.call()] {args_dict =}"))

        args = sig.bind(**args_dict).args

        function_response = self.fn(*args) #type: ignore

        # 构建函数结果消息对象
        return {
            "role": "tool",
            "tool_call_id": tool_call.id,
            "content": msgspec.json.encode(function_response).decode('utf-8')
        }

    def choice(self) -> ChatCompletionNamedToolChoiceParam:
        function_name = self.fn.__name__
        return {
            "type": "function",
            "function": {
                "name": function_name
            }
        }

def make_tool(fn: Callable[P, R]) -> MakeTool[P, R]:
    ''' Create a tool available for AI from a function '''
    return MakeTool(fn)
